# Copyright (c) 2010-2011 Outpost Embedded, LLC
# This file is part of the pydbusclone software package.
#
# pydbusclone is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 2 as published by the
# Free Software Foundation.
#
# A copy of the license has been included in the COPYING file.

from threading import RLock

import dbus

from dbusclone.base import (
  DBusObject,
  PropertiesObject,
  DBusObjectSet,
)
from dbusclone.util import CallbackList


class HalManager(DBusObject):
    devices = None

    def __init__(self, **kwargs):
        super(HalManager, self).__init__(
          bus_name = 'org.freedesktop.Hal',
          object_path = '/org/freedesktop/Hal/Manager',
          **kwargs
        )

        self.devices = DBusObjectSet(
          bus = self.bus,
          object_cls = HalDevice,
          ready_callback = self._handle_devices_ready,
          default_error_handler = self.default_error_handler,
        )

        self.connect_to_signal('DeviceAdded', self.devices.add_object)
        self.connect_to_signal('DeviceRemoved', self.devices.remove_object)
        self.call('GetAllDevices', reply_handler = self.devices.add_initial_objects)

    def get_proxy_object(self):
        obj = super(HalManager, self).get_proxy_object()
        return dbus.Interface(obj, 'org.freedesktop.Hal.Manager')

    def is_ready(self):
        if not super(HalManager, self).is_ready():
            return False
        return ((self.devices is not None) and self.devices.is_ready())

    def _handle_devices_ready(self, devices):
        self.maybe_ready()


class HalDevice(DBusObject):
    properties = None

    def __init__(self, **kwargs):
        super(HalDevice, self).__init__(
          bus_name = 'org.freedesktop.Hal',
          **kwargs
        )

        self.properties = HalDeviceProperties(
          bus = self.bus,
          parent = self,
          object_path = self.object_path,
          ready_callback = self._handle_properties_ready,
          default_error_handler = self.default_error_handler,
        )

    def get_proxy_object(self):
        obj = super(HalDevice, self).get_proxy_object()
        return dbus.Interface(obj, 'org.freedesktop.Hal.Device')

    def is_ready(self):
        if not super(HalDevice, self).is_ready():
            return False
        return (self.properties is not None) and self.properties.is_ready()

    def _handle_properties_ready(self, properties):
        self.maybe_ready()

    def has_capability(self, cap):
        try:
            caps = self.properties.get('info.capabilities')
        except KeyError:
            caps = ()
        return (cap in caps)


class HalDeviceProperties(PropertiesObject, DBusObject):
    _properties = None
    _lock = None

    def __init__(self, **kwargs):
        self._lock = RLock()

        super(HalDeviceProperties, self).__init__(
          bus_name = 'org.freedesktop.Hal',
          **kwargs
        )

        self.connect_to_signal(
          'PropertyModified',
          self._handle_property_modified,
        )
        self.call(
          'GetAllProperties',
          reply_handler = self._handle_get_all_properties_reply_init,
        )

    def get_proxy_object(self):
        obj = super(HalDeviceProperties, self).get_proxy_object()
        return dbus.Interface(obj, 'org.freedesktop.Hal.Device')

    def is_ready(self):
        if not super(HalDeviceProperties, self).is_ready():
            return False
        return (self._properties is not None)

    def _handle_get_all_properties_reply_init(self, properties):
        self._properties = properties
        self.maybe_ready()

    def _handle_property_modified(self, num_changes, changes):
        for change in changes:
            name, added, removed = change
            if removed:
                self._property_removed(name)
            else:
                self.call(
                  'GetProperty',
                  name,
                  reply_handler = self._make_handle_get_property_reply(name),
                )

    def _make_handle_get_property_reply(self, name):
        def _handle_get_property_reply(value):
            self._property_changed(name, value)
        return _handle_get_property_reply

    def _property_changed(self, name, value):
        if self._properties is None:
            return

        with self._lock:
            self._properties[name] = value
            properties = dict(self._properties)

        if self.is_ready():
            self.changed_callbacks.call(
              self.parent,
              [name],
              properties,
            )

    def _property_removed(self, name):
        if self._properties is None:
            return

        with self._lock:
            del self._properties[name]
            properties = dict(self._properties)

        if self.is_ready():
            self.changed_callbacks.call(
              self.parent,
              [name],
              properties,
            )

    def getdict(self, names = None):
        with self._lock:
            if names is None:
                return dict(self._properties)
            else:
                return dict(
                  (name, self._properties[name])
                  for name in names
                  if name in self._properties
                )

    def get(self, name, default = None):
        with self._lock:
            if default is not None:
                return self._properties.get(name, default)
            return self._properties[name]

    def set(name, value, reply_handler, error_handler):
        self.call(
          'SetProperty',
          name,
          value,
          reply_handler = reply_handler,
          error_handler = error_handler,
        )
